// Adaptation of old VIP fiji
// Now adapted for colour texture maps
/*
**
** fiji.c - F(i,j) illustrator
**
** Produces 2-D projection of shaded 3-D luminance surface F(i,j)
**
** usage:   fiji [-p] NOT perspective
**               [-x multiplier for grid scaling] 
**		 [-g amplitude_gain]
**		 [-v thetax]
**		 [-w thetaz] 
**		 [-l light_thetax] 
**		 [-m light_thetaz]  
**		 [-o focal length]
**		 [-d camera distance]
**		 [-t texture map file]
**		 in  [texture map file]
**
** John Robinson
*/



#include "picture.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

const float LUMINANCE_MULTIPLIER = 230;
const float AMBIENT = 20.0;
/* Biggest image we will handle */
const int BIGIMAGE = 1024;	

int ZGAIN = 16;
int XMULT = 1;
int YMULT = 1;
int focallength = 200;
int camdistance = 200;
int output_xsiz, output_ysiz;
float XPLUS = 0;
float YPLUS = 0;

int **outbuffer;
int outimage;
int	**inbuffer;	/* pointer to buffer of input image */
int	inimage;	/* input image number */
int maxval, minval;
picture_of_int *inpic;
picture_of_int *outpic;

/* Parameters to define angle of view */
const float DEGRAD = (float)3.141592/180;
float thetax = 0.0;
float thetaz = 0.0;
float light_thetax = -40.0;
float light_thetaz = -30.0;

int perspective_flag = 1;
int texture_flag = 0;
char in_filename[64];
char texture_filename[64];

// My implementation of getopt to can run on Windows
int optind = 1;
char *optarg = 0;
int my_getopt(int argc, char **argv, char *str) {
	if (optind == argc)
		return EOF;
	if (argv[optind][0] == '-') {
		char *p = str;
		while (*p) {
			if (*p == ':') { p++; continue; }
			if (*p == argv[optind][1]) {
				if (*(p+1) == ':') {
					optarg = argv[optind+1];
					optind += 2;
					return (int) *p;
					}
				else {
					optarg = 0; 
					optind++;
					return (int) *p;
					}
				}
			p++;
			}
		}
	return EOF;
	}

void handle_args(int argc, char *argv[])
    {
    int c;
    while ((c = my_getopt(argc, argv, "pg:v:w:l:m:x:d:o:t:")) != EOF)
        {
        switch (c)
            {
            case 'p':
                perspective_flag = 0;
                break;
            case 'v':
                thetax = (float) atof(optarg);
                break;
            case 'w':
                thetaz = (float) atof(optarg);
                break;
	    case 'x':
		XMULT = YMULT = atoi(optarg);
		break;
	    case 'g':
		ZGAIN = atoi(optarg);
		break;
	    case 'o':
		focallength = atoi(optarg);
		break;
	    case 'd':
		camdistance = atoi(optarg);
		break;
	    case 't':
		texture_flag = 1;
		strcpy(texture_filename,optarg);
		break;
            case 'l':
                light_thetax = (float) atof(optarg);
                break;
            case 'm':
                light_thetaz = (float) atof(optarg);
		break;
            }
        }
    if (((c = (argc - optind)) != 1)&&(c != 2))
	{
	fprintf(stderr,"Usage: fiji [options] depth_file [texture_file]\n");
	fprintf(stderr,"Options:\n");
	fprintf(stderr,"\t[-p] DON'T render in perspective\n");
	fprintf(stderr,"\t[-x multiplier for grid scaling] \n");
	fprintf(stderr,"\t[-g amplitude_gain]\n");
	fprintf(stderr,"\t[-v thetax]\n");
	fprintf(stderr,"\t[-w thetaz] \n");
	fprintf(stderr,"\t[-l light_thetax]\n"); 
	fprintf(stderr,"\t[-m light_thetaz]  \n");
	fprintf(stderr,"\t[-o focal length]\n");
	fprintf(stderr,"\t[-d camera distance]\n");
	fprintf(stderr,"\t[-t texture_file]\n");
	exit(1);
	}
    strcpy(in_filename,argv[optind++]);
    if (c == 2) {
	texture_flag = 1;
	strcpy(texture_filename,argv[optind]);
	}
    }

void create_clip_im() {
	outpic = new picture_of_int(output_ysiz, output_xsiz, 0);
	outbuffer = outpic->bufptr();
	*outpic = 0;
	}


int	nrows, ncols;	/* number of image rows and columns */
int get_image_and_range(char *infile) {
	colour_picture colinpic(infile);
	if (colinpic.ismonochrome())
		inpic = new picture_of_int(colinpic.mono);
	else
		{
		// Stack up colours into one picture_of_int
		int minval = colinpic.min();
		if (minval < 0) {
			colinpic.red -= minval;
			colinpic.green -= minval;
			colinpic.blue -= minval;
			}
		int maxval = colinpic.max();
		if (maxval > 255) {
			colinpic.red *= 255;
			colinpic.green *= 255;
			colinpic.blue *= 255;
			colinpic.red /= maxval;
			colinpic.green /= maxval;
			colinpic.blue /= maxval;
			}
		colinpic.red *= 256*256;
		colinpic.green *= 256;
		inpic = new picture_of_int(colinpic.blue);
		*inpic += colinpic.red;
		*inpic += colinpic.green;
		}
	nrows = inpic->nrows();
	ncols = inpic->ncols();
	inbuffer = inpic->bufptr();
	maxval = inpic->max();
	minval = inpic->min();
	return colinpic.ismonochrome();
}
void remove_image() {
	delete inpic;
	}


struct gridquad
	{
	int i, j;	/* Indices in array, for use after sorting */
	float depth;
	} *gridquad, *pgq;
struct grid
	{
	float coords[3];
	int overlayfunction;
	int overlayline;
	float transcoords[3];
	float luminance;
	} *grid, *pgrid;

/*
** Comparison routine for quicksorting
*/
int lowfirst(const void *p1, const void *p2){
	const struct gridquad *pg1 = (const struct gridquad *)p1;
	const struct gridquad *pg2 = (const struct gridquad *)p2;
	if (pg1->depth > pg2->depth)
		return(1);
	else if (pg2->depth > pg1->depth)
		return(-1);
	return(0);
	}



/* Transform coordinates by rotation */
void transform(float locthetaz,float locthetax,int perspective)
	{
	// cout << "transform(" << thetaz << "," << thetax << "," << perspective << ")\n";
	float sinthetaz, costhetaz;
	float sinthetax, costhetax;
	float xsines[BIGIMAGE];
	float xcosines[BIGIMAGE];
	float ysines[BIGIMAGE];
	float ycosines[BIGIMAGE];
	float newx,newy,newz;
	float outx,outy,outz;
	int i,j;
	sinthetaz = sin(locthetaz);
	costhetaz = cos(locthetaz);
	sinthetax = sin(locthetax);
	costhetax = cos(locthetax);
	/*
	** First do rotation about z. Use fact that x and y start life as
	** a regular grid
	*/
	pgrid = grid;
	for (i = 0; i < ncols; i++)
		{
		xsines[i] = pgrid->coords[0]*sinthetaz;
		xcosines[i] = pgrid->coords[0]*costhetaz;
		pgrid++;
		}
	pgrid = grid;
	for (i = 0; i < nrows; i++)
		{
		ysines[i] = pgrid->coords[1]*sinthetaz;
		ycosines[i] = pgrid->coords[1]*costhetaz;
		pgrid += ncols;
		}
	pgrid = grid;
	for (i = 0; i < nrows; i++)
	    {
	    for (j = 0; j < ncols; j++, pgrid++)
		{
		newx = xcosines[j] - ysines[i];
		newy = xsines[j] + ycosines[i];
		newz = pgrid->coords[2];
		outx = newx;
		outy = newz*sinthetax + newy*costhetax;
		outz = newz*costhetax - newy*sinthetax;
		if (perspective) {
			outx *= focallength/(camdistance-outz);
			outy *= focallength/(camdistance-outz);
			}
		pgrid->transcoords[0] = XMULT*outx;
		pgrid->transcoords[1] = YPLUS+YMULT*outy;
		pgrid->transcoords[2] = outz;
		}
	    }
	}

struct {
	int line;
	int xs, xe;
	int ls, le;
	} polyrast[BIGIMAGE];

		
int polyplot(int x,int y,int l)
	{
	int i;
	for (i = 0; i < BIGIMAGE; i++)
		{
		if (polyrast[i].line == y)
			{
			if (x < polyrast[i].xs)
				{
				polyrast[i].xs = x;
				polyrast[i].ls = l;
				}
			else if (x > polyrast[i].xe)
				{
				polyrast[i].xe = x;
				polyrast[i].le = l;
				}
			return(0);
			}
		else if (polyrast[i].line == -1)
			{
			polyrast[i].line = y;
			polyrast[i].xs = x;
			polyrast[i].xe = x;
			polyrast[i].ls = l;
			polyrast[i].le = l;
			return(0);
			}
		}
	return(-1);
	}
void xyplot(int **buf, int xs, int xe, int ys, int ye, int ls, int le) {
	register int x = (xs + xe)>>1;
	register int l;
	register int y = (ys + ye)>>1;
	if ((y == ys)||(y == ye))
		return;
	if (ls == le)
		l = ls;
	else 
		l = (ls>>1) + (le>>1);	// Works because 7 bit precision
					// means bottom bits are blanked
	buf[y][x] = l;
	l &= 0xfefefefeL;
	polyplot(x,y,l);
	xyplot(buf, xs, x, ys, y, ls, l);
	xyplot(buf, x, xe, y, ye, l, le);
	}
		
void xplot(int *buf, int xs, int xe, int ls, int le) {
	if (ls == le) {
		register int *p = buf + xs + 1;
		register int cnt = xe - xs - 1;
		while(cnt--)
			*p++ = ls;
		return;
		}
	int x = (xs + xe)>>1;
	int l = (ls>>1) + (le>>1);	// Works because 7 bit precision
					// means bottom bits are blanked
	buf[x] = l;
	l &= 0xfefefefeL;		// Continue 7 bit precision
	if (x > xs+1)
		xplot(buf, xs, x, ls, l);
	if (xe > x+1)
		xplot(buf, x, xe, l, le);
	}

void polygoninit() {
	for (int i = 0; i < BIGIMAGE; i++)
		polyrast[i].line = -1;	
	}

void polygon (picture_of_int& pic, int *x, int *y, int *l)
	{
	int i;
	int ll[4];
	// If any part of polygon is outside limits, don't draw
	for (i = 0; i < 4; i++) {
		if ((x[i] < 0)||(x[i] >= pic.ncols())||
			(y[i] < 0)||(y[i] >= pic.nrows()))
				return;
		}
	for (i = 0; i < BIGIMAGE; i++) {
		if (polyrast[i].line < 0)
			break;
		polyrast[i].line = -1;	
		}
	for (i = 0; i < 4; i++)
		ll[i] = l[i]&0xfefefefeL;	// All done to 7 bit precision
	for (i = 0; i < 4; i++) {
		pic[y[i]][x[i]] = ll[i];
		polyplot(x[i],y[i],ll[i]);
		}
	xyplot(pic.bufptr(),x[0],x[1],y[0],y[1],ll[0],ll[1]);
	xyplot(pic.bufptr(),x[1],x[2],y[1],y[2],ll[1],ll[2]);
	xyplot(pic.bufptr(),x[2],x[3],y[2],y[3],ll[2],ll[3]);
	xyplot(pic.bufptr(),x[3],x[0],y[3],y[0],ll[3],ll[0]);
	for (i = 0; i < BIGIMAGE; i++) {
		if (polyrast[i].line < 0)
			break;
		if (polyrast[i].xe > polyrast[i].xs + 1)
			xplot(pic.rowptr(polyrast[i].line),
				polyrast[i].xs, polyrast[i].xe,
				polyrast[i].ls, polyrast[i].le);
		}
	}
			
void get_normal(float *c1,float *c2,float *c3,float *c4,float *result)
	{
	float nx,ny,nz;
	float ux,uy,uz;
	float vx,vy,vz;
	float modulus;
	ux = c1[0] - c3[0];
	uy = c1[1] - c3[1];
	uz = c1[2] - c3[2];

	vx = c2[0] - c4[0];
	vy = c2[1] - c4[1];
	vz = c2[2] - c4[2];

	/* Cross product to find normal */
	nx = uy*vz - uz*vy;
	ny = - (ux*vz - uz*vx);
	nz = ux*vy - uy*vx;
	modulus = sqrt(nx*nx + ny*ny + nz*nz);
	if (nz < 0)
		nz = -nz;
	*result = AMBIENT + (LUMINANCE_MULTIPLIER*nz)/modulus;
	}

int main(int argc,char *argv[])
	{
	int i, j;
	int texture_is_monochrome = 1;

    handle_args(argc,argv);
    thetax *= (const float) DEGRAD;
    thetaz *= (const float) DEGRAD;
    light_thetax *= (const float) DEGRAD;
    light_thetaz *= (const float) DEGRAD;
    get_image_and_range(in_filename);
    grid = new struct grid[nrows*ncols];
    gridquad = new struct gridquad[(nrows-1)*(ncols-1)];
	polygoninit();

/* Generate values for z at grid points */

	pgrid = grid;
	for (i = 0; i < nrows; i++)
		{
		for (j = 0; j < ncols; j++)
			{
			pgrid->coords[0] =  (j - (float)((ncols-1)/2.0));
			pgrid->coords[1] =  (i - (float)((nrows-1)/2.0));
     pgrid->coords[2] = ((float)(inbuffer[i][j]-minval)*ZGAIN)/(maxval-minval);
			pgrid++;
			}
		}
	remove_image( );
	output_xsiz = ncols;
	output_ysiz = nrows;
	if (texture_flag)
		texture_is_monochrome = get_image_and_range(texture_filename);
	else	
	/* Deal with lighting */
	/*  transform all grid points */
		transform(light_thetaz,light_thetax,0);
	/* Now work out normals */
	pgrid = grid;
	for (i = 0; i < nrows; i++)
		{
		for (j = 0; j < ncols; j++)
			{
			if ((i==0)||(j==0)||(i==nrows-1)||(j==ncols-1))
				{
				pgrid->luminance = 0.0;
				pgrid++;
				continue;
				}
			if (texture_flag) 
				pgrid->luminance = (float) inbuffer[i][j];
			else
				get_normal((pgrid-1)->transcoords,
				(pgrid-ncols)->transcoords,
				(pgrid+1)->transcoords,
				(pgrid+ncols)->transcoords,
				&pgrid->luminance);
			pgrid++;
			}
		}
	create_clip_im(); // So this sets size of output pic
	XPLUS = (float)(ncols/2);
	YPLUS = (float)(nrows/2);
	outpic->show("Fiji picture");
	int pointerx = -1;
	int pointery = -1;
	int thistime = 1;
	int pointingleft = 0;
	int pointingright = 0;
	int tophalf = 0;	// For left point and x movement
	while(1) {
		if (outpic->closerequested())
			break;
		if (!thistime) {
			if (outpic->pointleftbutton()) {
			    if (pointingleft && ((outpic->pointx() != pointerx)||(outpic->pointy() != pointery))) {
				int pointxoffset = pointerx - outpic->pointx();
				int pointyoffset = pointery - outpic->pointy();
				pointerx = outpic->pointx();
				pointery = outpic->pointy();
				if (tophalf)
				  thetaz-=pointxoffset*360*DEGRAD/(output_xsiz);
				else 
				  thetaz+=pointxoffset*360*DEGRAD/(output_xsiz);
				thetax -= pointyoffset*180*DEGRAD/(output_ysiz);
				thistime = 1;
				}
			    else if (!pointingleft) {
				pointingleft = 1;
				pointerx = outpic->pointx();
				pointery = outpic->pointy();
				if (pointery < output_ysiz/2)
					tophalf = 1;
				else
					tophalf = 0;
				}
			    }
			else
				pointingleft = 0;
			if (outpic->pointrightbutton()) {
			    if (pointingright && ((outpic->pointx() != pointerx)||(outpic->pointy() != pointery))) {
				int pointxoffset = pointerx - outpic->pointx();
				int pointyoffset = pointery - outpic->pointy();
				pointerx = outpic->pointx();
				pointery = outpic->pointy();
				camdistance -= pointyoffset*200/output_ysiz;
				XPLUS -= pointxoffset;
				thistime = 1;
				}
			    else if (!pointingright) {
				pointingright = 1;
				pointerx = outpic->pointx();
				pointery = outpic->pointy();
				}
			    }
			else
				pointingright = 0;
			}
		if (!thistime)
			continue;	// Don't refresh
		thistime--;
		/* Retransform all grid points */
		transform(thetaz,thetax,perspective_flag);
		/* Put into gridquad arrays, then order, lowest first */
		pgq = gridquad;
		for (i = 0; i < nrows-1; i++)
	    	{
	    	for (j = 0; j < ncols-1; j++)
			{
			pgrid = grid + i*ncols + j;
			pgq->i = i;
			pgq->j = j;
			pgq->depth = pgrid->transcoords[2] +
					(pgrid+1)->transcoords[2] +
					(pgrid+ncols+1)->transcoords[2] +
					(pgrid+ncols)->transcoords[2];
			pgq->depth /= 4.0;
			pgq++;
			}
	    	}
		qsort(gridquad,(ncols-1)*(nrows-1),sizeof(struct
			gridquad), lowfirst);
		/* Now have ordered list of quads. Simply draw them in order */
		pgq = gridquad;
		for (i = 0; i < nrows-1; i++)
	    	{
	    	for (j = 0; j < ncols-1; j++)
			{
			int l[4];	/* Luminances at the vertices */
			int vx[4];
			int vy[4];
			struct grid *p;
			p = grid + (ncols*pgq->i) + pgq->j;
			l[0] = (int) p->luminance;
			vx[0] = (int)(p->transcoords[0] + XPLUS);
			vy[0] = (int)(p->transcoords[1]);
			p++;
			l[1] = (int) p->luminance;
			vx[1] = (int)(p->transcoords[0] + XPLUS);
			vy[1] = (int)(p->transcoords[1]);
			p += ncols;
			l[2] = (int) p->luminance;
			vx[2] = (int)(p->transcoords[0] + XPLUS);
			vy[2] = (int)(p->transcoords[1]);
			p--;
			l[3] = (int) p->luminance;
			vx[3] = (int)(p->transcoords[0] + XPLUS);
			vy[3] = (int)(p->transcoords[1]);
			polygon(*outpic,vx,vy,l);
			pgq++;
			}
	    	}
	if (texture_is_monochrome)
		outpic->reshow();
	else
		outpic->reshowascolour();
	*outpic = 0;
	}
	return 0;
	}
